void GSAHWEnable(void);
u8 CheckGSAButton(void);
u8 CheckGSASwitch(void);
void CheckUSBReady(u8 flag);
void ReadUSB(u8 *data);
void WriteUSB(u8 *data);
void hook1(u32 addr);
void remap1(u32 addr, u16 data);
void remap2(u32 addr, u16 data);
void remap3(u32 addr, u16 data);


#define USB_READY        0x0100
#define GSA_BUTTON       0x0400
//#define USB_CONNECT      0x0800
#define GSA_SWITCH       0xF000
#define GSA_DISABLE      0x0000
#define GSA_ENABLE       0x0FF0
#define GSA_ROMSWAP      0x0001
#define GSA_WRITE        0x0020
#define GSA_READ         0x0040

#define REG_GSAFLASH     *(u16*)0x0800009C
#define REG_HWIN         *(u16*)0x09FE01C0
#define REG_HWOUT        *(u16*)0x09FE01E0
#define REG_GSAENABLE    *(u16*)0x09FFE140

#define REG_REMAP1LOW    *(u16*)0x09FE0000
#define REG_REMAP1HI     *(u16*)0x09FE0020
#define REG_REMAP1VAL    *(u16*)0x09FE0040
#define REG_REMAP2LOW    *(u16*)0x09FE0080
#define REG_REMAP2HI     *(u16*)0x09FE00A0
#define REG_REMAP2VAL    *(u16*)0x09FE00C0
#define REG_REMAP3LOW    *(u16*)0x09FE0100
#define REG_REMAP3HI     *(u16*)0x09FE0120
#define REG_REMAP3VAL    *(u16*)0x09FE0060


//GSAHWEnable enables flash write, and master flash enable
void GSAHWEnable() {
	REG_GSAFLASH = GSA_WRITE;
	REG_GSAENABLE = GSA_ENABLE;
}
//CheckGSAButton returns true or false, depending on button state: true = pressed
u8 CheckGSAButton() {
	if (REG_HWIN & GSA_BUTTON) return 0;
	return 1;
}
//CheckGSASwitch returns true or false, depending on switch state: true = ON
u8 CheckGSASwitch() {
	if ((REG_HWIN & GSA_SWITCH) == GSA_SWITCH) return 1;
	return 0;
}

//CheckUSBReady waits until the USB_READY flag is set to the specified value
void CheckUSBReady(u8 flag) {
	//while ((REG_HWIN & USB_READY) == (flag << 0x08));
	u32 i;
	for (i = 0; i < 0x2000; i++) {
		if ((REG_HWIN & USB_READY) != (flag << 0x08)) return;
	}
	print("\n USB ERROR!\n  Transfer Timeout\n  USB_READY: 0x%02X\n\n Please Reset Your GBA Now", flag);
	while(1);
}
//ReadUSB reads 8 bytes into buffer from USB
//buffer is a pointer to an 8-byte u8 array
void ReadUSB(u8 *buffer) {
	u8 i;
	for (i = 0; i < 4; i++) {
		CheckUSBReady(0);
		buffer[i<<1] = (u8)REG_HWIN;
		REG_HWOUT = 0x0100;
		CheckUSBReady(1);
		buffer[(i<<1)+1] = (u8)REG_HWIN;
		REG_HWOUT = 0x0000;
	}
}
//WriteUSB sends 8 bytes from buffer to USB
//buffer is a pointer to an 8-byte u8 array
void WriteUSB(u8 *buffer) {
	u8 i;
	for (i = 0; i < 4; i++) {
		CheckUSBReady(0);
		REG_HWOUT = buffer[i<<1];
		REG_HWOUT = buffer[i<<1] | USB_READY;
		CheckUSBReady(1);
		REG_HWOUT = buffer[(i<<1)+1] | USB_READY;
		REG_HWOUT = buffer[(i<<1)+1];
	}
}
//remap ROM address, basic "ROM write"
void remap1(u32 addr, u16 data) {
	addr >>= 1;
	REG_REMAP1HI = addr >> 0x10;
	REG_REMAP1LOW = addr & 0xFFFF;
	REG_REMAP1VAL = data;
}
void remap2(u32 addr, u16 data) {
	addr >>= 1;
	REG_REMAP2HI = addr >> 0x10;
	REG_REMAP2LOW = addr & 0xFFFF;
	REG_REMAP2VAL = data;
}
void remap3(u32 addr, u16 data) {
	addr >>= 1;
	REG_REMAP3HI = addr >> 0x10;
	REG_REMAP3LOW = addr & 0xFFFF;
	REG_REMAP3VAL = data;
}
